@@ -2,54 +2,43 @@ import asyncio
 import contextvars
 import dataclasses
 import inspect
+import logging
 import traceback
 from typing import Any
 from typing import Callable
 from typing import List
 from typing import Optional
 
-import orjson
-
 from autoblocks._impl import global_state
+from autoblocks._impl.context_vars import TestCaseRunContext
+from autoblocks._impl.context_vars import test_case_run_context_var
 from autoblocks._impl.testing.models import BaseTestCase
 from autoblocks._impl.testing.models import BaseTestEvaluator
-from autoblocks._impl.testing.models import Evaluation
+from autoblocks._impl.testing.models import TestCaseContext
+from autoblocks._impl.testing.util import serialize
+from autoblocks._impl.testing.util import serialize_test_case
+from autoblocks._impl.testing.util import yield_test_case_contexts_from_test_cases
 from autoblocks._impl.util import AutoblocksEnvVar
-from autoblocks._impl.util import gather_with_max_concurrency
+from autoblocks._impl.util import all_settled
 
-# Globals
-_cli_server_address: Optional[str] = None
+log = logging.getLogger(__name__)
+
+test_case_semaphore_registry: dict[str, asyncio.Semaphore] = {}  # test_id -> semaphore
+evaluator_semaphore_registry: dict[str, dict[str, asyncio.Semaphore]] = {}  # test_id -> evaluator_id -> semaphore
 
 
 def cli() -> str:
     """
     Returns the CLI server address, which is required to send results and errors to the CLI.
     """
-    global _cli_server_address
-    if not _cli_server_address:
-        cli_server_address = AutoblocksEnvVar.CLI_SERVER_ADDRESS.get()
-        if not cli_server_address:
-            raise RuntimeError(
-                "Autoblocks tests must be run within the context of the testing CLI.\n"
-                "Make sure you are running your test command with:\n"
-                "$ npx autoblocks testing exec -- <your test command>"
-            )
-        _cli_server_address = cli_server_address
-    return _cli_server_address
-
-
-def orjson_default(o: Any) -> Any:
-    if hasattr(o, "model_dump_json") and callable(o.model_dump_json):
-        # pydantic v2
-        return orjson.loads(o.model_dump_json())
-    elif hasattr(o, "json") and callable(o.json):
-        # pydantic v1
-        return orjson.loads(o.json())
-    raise TypeError
-
-
-def serialize(x: Any) -> Any:
-    return orjson.loads(orjson.dumps(x, default=orjson_default))
+    cli_server_address = AutoblocksEnvVar.CLI_SERVER_ADDRESS.get()
+    if not cli_server_address:
+        raise RuntimeError(
+            "Autoblocks tests must be run within the context of the testing CLI.\n"
+            "Make sure you are running your test command with:\n"
+            "$ npx autoblocks testing exec -- <your test command>"
+        )
+    return cli_server_address
 
 
 async def send_error(
@@ -73,108 +62,127 @@ async def send_error(
     )
 
 
-async def evaluate_output(
+async def run_evaluator_unsafe(
     test_id: str,
-    test_case: BaseTestCase,
+    test_case_ctx: TestCaseContext,
     output: Any,
     evaluator: BaseTestEvaluator,
 ) -> None:
-    evaluation: Optional[Evaluation] = None
-
-    if inspect.iscoroutinefunction(evaluator.evaluate_test_case):
-        try:
-            evaluation = await evaluator.evaluate_test_case(test_case, output)
-        except Exception as err:
-            await send_error(
-                test_id=test_id,
-                test_case_hash=test_case._cached_hash,
-                evaluator_id=evaluator.id,
-                error=err,
-            )
-    else:
-        try:
+    """
+    This is suffixed with _unsafe because it doesn't handle exceptions.
+    Its caller will catch and handle all exceptions.
+    """
+    async with evaluator_semaphore_registry[test_id][evaluator.id]:
+        if inspect.iscoroutinefunction(evaluator.evaluate_test_case):
+            evaluation = await evaluator.evaluate_test_case(test_case_ctx.test_case, output)
+        else:
             ctx = contextvars.copy_context()
             evaluation = await global_state.event_loop().run_in_executor(
                 None,
                 ctx.run,
                 evaluator.evaluate_test_case,
-                test_case,
+                test_case_ctx.test_case,
                 output,
             )
-        except Exception as err:
-            await send_error(
-                test_id=test_id,
-                test_case_hash=test_case._cached_hash,
-                evaluator_id=evaluator.id,
-                error=err,
-            )
-
-    if evaluation is None:
-        return
 
     await global_state.http_client().post(
         f"{cli()}/evals",
         json=dict(
             testExternalId=test_id,
-            testCaseHash=test_case._cached_hash,
+            testCaseHash=test_case_ctx.hash(),
             evaluatorExternalId=evaluator.id,
             score=evaluation.score,
             threshold=dataclasses.asdict(evaluation.threshold) if evaluation.threshold else None,
+            metadata=evaluation.metadata,
         ),
     )
 
 
-async def run_test_case(
+async def run_evaluator(
     test_id: str,
-    test_case: BaseTestCase,
-    evaluators: List[BaseTestEvaluator],
-    fn: Callable[[BaseTestCase], Any],
-    max_evaluator_concurrency: int,
+    test_case_ctx: TestCaseContext,
+    output: Any,
+    evaluator: BaseTestEvaluator,
 ) -> None:
-    output = None
+    try:
+        await run_evaluator_unsafe(
+            test_id=test_id,
+            test_case_ctx=test_case_ctx,
+            output=output,
+            evaluator=evaluator,
+        )
+    except Exception as err:
+        await send_error(
+            test_id=test_id,
+            test_case_hash=test_case_ctx.hash(),
+            evaluator_id=evaluator.id,
+            error=err,
+        )
 
-    if inspect.iscoroutinefunction(fn):
-        try:
-            output = await fn(test_case)
-        except Exception as err:
-            await send_error(
-                test_id=test_id,
-                test_case_hash=test_case._cached_hash,
-                evaluator_id=None,
-                error=err,
-            )
-    else:
-        try:
+
+async def run_test_case_unsafe(
+    test_id: str,
+    test_case_ctx: TestCaseContext,
+    fn: Callable[[BaseTestCase], Any],
+) -> Any:
+    """
+    This is suffixed with _unsafe because it doesn't handle exceptions.
+    Its caller will catch and handle all exceptions.
+    """
+    async with test_case_semaphore_registry[test_id]:
+        if inspect.iscoroutinefunction(fn):
+            output = await fn(test_case_ctx.test_case)
+        else:
             ctx = contextvars.copy_context()
-            output = await global_state.event_loop().run_in_executor(None, ctx.run, fn, test_case)
-        except Exception as err:
-            await send_error(
-                test_id=test_id,
-                test_case_hash=test_case._cached_hash,
-                evaluator_id=None,
-                error=err,
+            output = await global_state.event_loop().run_in_executor(
+                None,
+                ctx.run,
+                fn,
+                test_case_ctx.test_case,
             )
 
-    if output is None:
-        return
-
     await global_state.http_client().post(
         f"{cli()}/results",
         json=dict(
             testExternalId=test_id,
-            testCaseHash=test_case._cached_hash,
-            testCaseBody=serialize(test_case),
+            testCaseHash=test_case_ctx.hash(),
+            testCaseBody=serialize_test_case(test_case_ctx.test_case),
             testCaseOutput=serialize(output),
         ),
     )
+    return output
+
+
+async def run_test_case(
+    test_id: str,
+    test_case_ctx: TestCaseContext,
+    evaluators: List[BaseTestEvaluator],
+    fn: Callable[[BaseTestCase], Any],
+) -> None:
+    token = test_case_run_context_var.set(TestCaseRunContext(test_id=test_id, test_case_hash=test_case_ctx.hash()))
+    try:
+        output = await run_test_case_unsafe(
+            test_id=test_id,
+            test_case_ctx=test_case_ctx,
+            fn=fn,
+        )
+    except Exception as err:
+        await send_error(
+            test_id=test_id,
+            test_case_hash=test_case_ctx.hash(),
+            evaluator_id=None,
+            error=err,
+        )
+        return
+    finally:
+        test_case_run_context_var.reset(token)
 
     try:
-        await gather_with_max_concurrency(
-            max_evaluator_concurrency,
+        await all_settled(
             [
-                evaluate_output(
+                run_evaluator(
                     test_id=test_id,
-                    test_case=test_case,
+                    test_case_ctx=test_case_ctx,
                     output=output,
                     evaluator=evaluator,
                 )
@@ -184,32 +192,43 @@ async def run_test_case(
     except Exception as err:
         await send_error(
             test_id=test_id,
-            test_case_hash=test_case._cached_hash,
+            test_case_hash=test_case_ctx.hash(),
             evaluator_id=None,
             error=err,
         )
 
 
+def validate_test_suite_inputs(
+    test_id: str,
+    test_cases: List[BaseTestCase],
+    evaluators: List[BaseTestEvaluator],
+) -> None:
+    assert test_cases, f"[{test_id}] No test cases provided."
+    for test_case in test_cases:
+        assert isinstance(
+            test_case,
+            BaseTestCase,
+        ), f"[{test_id}] Test case {test_case} does not implement {BaseTestCase.__name__}."
+    for evaluator in evaluators:
+        assert isinstance(
+            evaluator,
+            BaseTestEvaluator,
+        ), f"[{test_id}] Evaluator {evaluator} does not implement {BaseTestEvaluator.__name__}."
+
+
 async def async_run_test_suite(
     test_id: str,
     test_cases: List[BaseTestCase],
     evaluators: List[BaseTestEvaluator],
     fn: Callable[[BaseTestCase], Any],
     max_test_case_concurrency: int,
-    max_evaluator_concurrency: int,
 ) -> None:
     try:
-        assert test_cases, f"[{test_id}] No test cases provided."
-        for test_case in test_cases:
-            assert isinstance(
-                test_case,
-                BaseTestCase,
-            ), f"[{test_id}] Test case {test_case} does not implement {BaseTestCase.__name__}."
-        for evaluator in evaluators:
-            assert isinstance(
-                evaluator,
-                BaseTestEvaluator,
-            ), f"[{test_id}] Evaluator {evaluator} does not implement {BaseTestEvaluator.__name__}."
+        validate_test_suite_inputs(
+            test_id=test_id,
+            test_cases=test_cases,
+            evaluators=evaluators,
+        )
     except Exception as err:
         await send_error(
             test_id=test_id,
@@ -219,20 +238,24 @@ async def async_run_test_suite(
         )
         return
 
+    # Initialize the semaphore registries
+    test_case_semaphore_registry[test_id] = asyncio.Semaphore(max_test_case_concurrency)
+    evaluator_semaphore_registry[test_id] = {
+        evaluator.id: asyncio.Semaphore(evaluator.max_concurrency) for evaluator in evaluators
+    }
+
     await global_state.http_client().post(f"{cli()}/start", json=dict(testExternalId=test_id))
 
     try:
-        await gather_with_max_concurrency(
-            max_test_case_concurrency,
+        await all_settled(
             [
                 run_test_case(
                     test_id=test_id,
-                    test_case=test_case,
+                    test_case_ctx=test_case_ctx,
                     evaluators=evaluators,
                     fn=fn,
-                    max_evaluator_concurrency=max_evaluator_concurrency,
                 )
-                for test_case in test_cases
+                for test_case_ctx in yield_test_case_contexts_from_test_cases(test_cases)
             ],
         )
     except Exception as err:
@@ -253,19 +276,26 @@ def run_test_suite(
     fn: Callable[[BaseTestCase], Any],
     # How many test cases to run concurrently
     max_test_case_concurrency: int = 10,
-    # How many evaluators to run concurrently on the result of a test case
-    max_evaluator_concurrency: int = 5,
+    # Deprecated arguments, but left for backwards compatibility
+    max_evaluator_concurrency: Optional[int] = None,
 ) -> None:
     global_state.init()
-    future = asyncio.run_coroutine_threadsafe(
+
+    if max_evaluator_concurrency is not None:
+        log.warning(
+            "`max_evaluator_concurrency` is deprecated and will be removed in a future release.\n"
+            "Its value is being ignored.\n"
+            "Set the `max_concurrency` attribute on the evaluator class instead.\n"
+            "See https://docs.autoblocks.ai/testing/sdks for more information."
+        )
+
+    asyncio.run_coroutine_threadsafe(
         async_run_test_suite(
             test_id=id,
             test_cases=test_cases,
             evaluators=evaluators,
             fn=fn,
             max_test_case_concurrency=max_test_case_concurrency,
-            max_evaluator_concurrency=max_evaluator_concurrency,
         ),
         global_state.event_loop(),
-    )
-    future.result()
+    ).result()
