@@ -5,8 +5,9 @@ import urllib.parse
 from enum import Enum
 from typing import Any
 from typing import Coroutine
-from typing import List
 from typing import Optional
+from typing import Tuple
+from typing import Union
 
 log = logging.getLogger(__name__)
 
@@ -21,6 +22,7 @@ class AutoblocksEnvVar(StrEnum):
     API_KEY = "AUTOBLOCKS_API_KEY"
     CLI_SERVER_ADDRESS = "AUTOBLOCKS_CLI_SERVER_ADDRESS"
     INGESTION_KEY = "AUTOBLOCKS_INGESTION_KEY"
+    PROMPT_SNAPSHOTS = "AUTOBLOCKS_PROMPT_SNAPSHOTS"
     TRACER_THROW_ON_ERROR = "AUTOBLOCKS_TRACER_THROW_ON_ERROR"
 
     def get(self) -> Optional[str]:
@@ -36,19 +38,22 @@ def encode_uri_component(s: str) -> str:
     return urllib.parse.quote(s, safe="()*!.'")
 
 
-async def gather_with_max_concurrency(
-    max_concurrency: int,
-    coroutines: List[Coroutine[Any, Any, Any]],
-) -> None:
+async def all_settled(coroutines: list[Coroutine[Any, Any, Any]]) -> Tuple[Union[BaseException, Any]]:
     """
-    Borrowed from https://stackoverflow.com/a/61478547
+    Runs all the coroutines in parallel and waits for all of them to finish,
+    regardless of whether they succeed or fail. Similar to Promise.allSettled.
     """
-    semaphore = asyncio.Semaphore(max_concurrency)
+    # mypy error:
+    # Returning Any from function declared to return "tuple[BaseException | Any]"  [no-any-return]
+    return await asyncio.gather(*coroutines, return_exceptions=True)  # type: ignore
 
-    async def sem_coro(coro: Coroutine[Any, Any, Any]) -> Any:
-        async with semaphore:
-            return await coro
 
-    # return_exceptions=True causes exceptions to be returned as values instead
-    # of propagating them to the caller. this is similar in behavior to Promise.allSettled
-    await asyncio.gather(*(sem_coro(c) for c in coroutines), return_exceptions=True)
+def get_running_loop() -> Optional[asyncio.AbstractEventLoop]:
+    """
+    Returns the currently-running event loop if there is one.
+    """
+    try:
+        # Raises a RuntimeError if no loop is running
+        return asyncio.get_running_loop()
+    except RuntimeError:
+        return None
